;;;-*- Mode: Lisp; Encoding: (win32:code-page :id 1251) -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Column header facility
;;;
;;; Linux/Mac OS TODO: Provide meaningful resize and select cursors

(in-package :ystok.grid)

;;  Maybe use gp:register-image-translation and gp:load-image instead?
(defconstant +column-header-image-width+ 10 "One image width")
(defconstant +column-header-image-height+ 9)

;; Area at column right where double-click causes for resizing to fit best.
(defparameter *column-header-second-press-tolerance* 10)

(defvar *column-header-external-image*
  #.(gp:read-external-image (merge-pathnames "images/columnhdr.bmp"
                                        (or *compile-file-truename* *load-truename*))
                            :transparent-color-index 13))  ; magenta

(defvar *column-header-image* nil "Images to be drawn on an ordinary column header")
(defvar *column-header-image-selected* nil "Images to be drawn on a selected column header")

(defun column-header-image (header)
  (if (or (null *column-header-image*) (gp:image-freed-p *column-header-image*))
      (setq *column-header-image* (gp:convert-external-image header
                                                       *column-header-external-image*))
      *column-header-image*))

(defun column-header-image-selected (header)
  (if (or (null *column-header-image-selected*)
          (gp:image-freed-p *column-header-image-selected*))
      (gp:with-graphics-state (header :background
                                      #-(and lispworks4 win32) :color_3dhighlight
                                      #+(and lispworks4 win32) 'win32:color_3dhighlight)
        (setq *column-header-image-selected* (gp:convert-external-image header
                                                       *column-header-external-image*)))
      *column-header-image-selected*))

(defvar *column-header-resize-cursor*
  #+(or win32 (not lispworks4)) (ywi:load-cursor-from-file
                                 #.(lw:current-pathname "images/ResizeWE.cur"))
  #-(or win32 (not lispworks4)) :h-double-arrow)
(defun column-header-resize-cursor-p (pane)
  (ywi:cursor-eql pane *column-header-resize-cursor*))
(defun column-header-resize-cursor (pane)
  (unless (column-header-resize-cursor-p pane)		; provide resize cursor
    (ywi:set-cursor pane *column-header-resize-cursor*)))

(defvar *column-header-select-cursor*
  #+(or win32 (not lispworks4)) (ywi:load-cursor-from-file
                                 #.(lw:current-pathname "images/tbarrow.cur"))
  #-(or win32 (not lispworks4)) :hand)
(defun column-header-select-cursor-p (pane)
  (ywi:cursor-eql pane *column-header-select-cursor*))
(defun column-header-select-cursor (pane)
  (unless (column-header-select-cursor-p pane)		; provide select cursor
    (ywi:set-cursor pane *column-header-select-cursor*)))

(defparameter *column-header-input-model*
  (nconc
   '((:motion                     column-header-motion)
     ((:motion :shift)            column-header-motion)
     ((:button-1 :motion)         column-header-button-1-motion)
     ;((:button-1 :motion :shift) column-header-button-1-motion)
     ((:button-1 :press)          column-header-button-1-press)
     ((:button-1 :press :control) column-header-control-button-1-press)
     ((:button-1 :press :shift)   column-header-shift-button-1-press)
     ((:button-1 :release)        column-header-button-1-release)
     ((:button-1 :second-press)	  component-button-1-second-press)
     ;((:button-1 :release :control) column-header-button-1-release)
     ;((:button-1 :release :shift)   column-header-button-1-release)
    )
   *component-input-model*))
                  
(defclass column-header-row (grid-component capi:output-pane)
 ((resize-press-x :initform nil)		; x where resize gesture starts
  (select-press-column :accessor select-press-column
                  :initform nil))		; column where select gesture starts
 (:default-initargs
  :display-callback '%column-header-on-display
  :input-model *column-header-input-model*
  :accepts-focus-p nil
  :background #-(and lispworks4 win32) :color_3dface
              #+(and lispworks4 win32) 'win32:color_3dface
  :foreground #-(and lispworks4 win32) :color_btntext
              #+(and lispworks4 win32) 'win32:color_btntext
  :visible-min-width 1 :visible-max-width nil
  :visible-max-height t ; :visible-min-height *default-column-header-height*
  :visible-border nil))
  ;:help-key #L"LEFT-CLICK on this header to select the entire column"))

(defmethod component-on-create :after ((header column-header-row))
  ;; Don't override min-height if it has been already set explicitely
  (capi:set-geometric-hint header :visible-min-height (+ (%font-height% header) 3) nil)
  ;(when (and *use-os-cursors* (zerop *column-header-resize-cursor*))
  #+(and win32 lispworks4)
  (when (zerop *column-header-resize-cursor*)
    (setq *column-header-select-cursor* (ywi:load-cursor-from-file
                                         (lw:current-pathname "images/tbarrow.cur"))
          *column-header-resize-cursor* (ywi:load-cursor-from-file
                                         (lw:current-pathname "images/ResizeWE.cur")))))
      ;(setq *use-os-cursors* nil))))	; since cursor files was not found

(defun %column-header-on-display (column-header left top width height)
 ;;; Args: left,top,width,height Visual coordinates
  ;; CAUTION: Can be invoked outside the top-level-hook handlers!
  (let* ((grid (capi:element-interface column-header))
         (scroll-x (scroll-x grid))
         (*background-cleared* t))
    (gp:with-graphics-translation (column-header (- scroll-x) 0)
      (draw-column-header grid column-header (+ left scroll-x) top width height))))

(defun column-header-redraw (grid &optional column)
 ;;; Args: column If given, redraw the specified column only
  (let ((column-header (column-header grid))
        (*background-cleared* nil))
    (gp:with-graphics-translation (column-header (- (scroll-x grid)) 0)
      (if column
          (draw-column-header grid column-header (column-left grid column)
                              ;(+ (column-left grid column)
                              ;   (if (left-frozen-column-p (column-number column)
                              ;                             (left-frozen-ncols grid))
                              ;       (scroll-x grid) 0))
                              0 (column-width column))
          (draw-column-header grid column-header)))))

(defun draw-column-header (grid port &optional left (top 0) width height)
 ;;; Render header contents from (left, top) to (left+width, top+height)
  ;; Args: left,top Internal coordinates.
  ;;       port	    Header itself or printer-port with appropriate transform applied.
  ;; NB: Looping is in internal coordinates
  ;(format t "~&draw-column-header transform: ~S" (gp:graphics-port-transform port))
  (declare (fixnum top)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let* ((column-header (column-header grid))
         (left (the fixnum (or left (scroll-x grid))))
         (width (the fixnum (or width (gp:port-width column-header))))
         (height (the fixnum (or height (gp:port-height column-header))))
         (right (the fixnum (+ left width)))
         (graphics-args (unless (eq port column-header)
                          (when-let (font (capi:simple-pane-font column-header))
                            (list :font font))))
         (columns (columns grid))
         (ncols (ncols grid))
         (left-frozen-ncols (left-frozen-ncols grid))
         (x 0))
    (declare (fixnum left width height right ncols left-frozen-ncols)
             (type simple-vector columns))
    (apply #'gp:clear-rectangle port 
           (translate-mask port nil :left left :top top :width width :height height))
    (when (and (= top 0) (border column-header))		; black line on top
      (gp:draw-line port left 0 right 0				;3dshadow
                    :foreground #1= #-(and lispworks4 win32) :color_windowtext
                                    #+(and lispworks4 win32) 'win32:color_windowtext))
    ;; Draw frozen columns:  left      
    (when (plusp left-frozen-ncols)
      (let ((scroll-x (scroll-x grid))
            (left-frozen-width (left-frozen-width grid)))
        (do (column
             ;(right (gp:port-width column-header))
             (next-x 0)
             (x scroll-x next-x)				; back to visial
             (col 0 (1+ col)))
            ((or (>= col left-frozen-ncols) (< right x)))
          (declare (fixnum next-x x col))
          (setq column (svref columns col)
                next-x (the fixnum (+ x (the fixnum (column-width column))))) ; cell right
          (when (< left next-x)
            (draw-cell grid column-header (setq column (svref columns col)) port x
                       graphics-args)))
        (setq left (the fixnum (max left (+ scroll-x left-frozen-width)))
              x left-frozen-width)))
    ;; Draw scrollable columns
    (do (column
         (next-x x)
         (col left-frozen-ncols (1+ col)))
        ((or (>= col ncols) (< right x))
         (when (< next-x right)		; fill background outside internal grid area
           (let ((y (the fixnum (1- (gp:port-height column-header)))))
             (declare (fixnum y))
             (gp:draw-line port next-x y right y :foreground #1#))))
      (setq column (svref columns col)
            next-x (the fixnum (+ x (the fixnum (column-width column))))) ; cell right
      (when (< left next-x)
        (draw-cell grid column-header column port x graphics-args))
      (setq x next-x))))

(defmethod draw-cell :before (grid (column-header column-header-row) column port x
                              &optional graphics-args)
 ;;; Draw 3d border around the cell
  ;; Args: x  Left internal coordinate of cell rectangle
  (declare (ignore graphics-args)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let* ((column-width (column-width column))
         (height (gp:port-height column-header))
         (x (the fixnum (or x (column-left grid column))))
         (top (the fixnum (if (border column-header) 1 0))))
    (declare (fixnum column-width height x top))
    (with-cell-mask (mask-left mask-width) grid column x column-width
      (when (<= mask-width 0)
        (return-from draw-cell))
      (unless port (setq port column-header))
      ;; Draw 3d border
      (cond ((column-selected column)
             #1=(apply #'gp:clear-rectangle port
                       (translate-mask port nil :left mask-left :top top
                                       :width mask-width :height (1- height)))
             (gp:draw-rectangle port (the fixnum (1+ mask-left)) 2
                                (the fixnum (- mask-width 2)) ;column-width?
                                (the fixnum (- height 3))
                                :filled t
                                :foreground
                                  #2= #-(and lispworks4 win32) :color_3dhighlight
                                      #+(and lispworks4 win32) 'win32:color_3dhighlight)
             (gp:draw-polygon port #3=(list (the fixnum (+ mask-left mask-width -1)) 
                                            top
                                            (the fixnum (+ mask-left mask-width -1))
                                            (the fixnum (1- height))
                                            (the fixnum (1- mask-left))
                                            (the fixnum (1- height)))
                              :foreground :white))
            (t
             (unless *background-cleared* #1#)
             (gp:draw-line port mask-left top (the fixnum (+ mask-left mask-width)) top
                           :foreground #2#)
             (gp:draw-polygon port #3#				 ; 3dshadow
                         :foreground #-(and lispworks4 win32) :color_windowtext
                                     #+(and lispworks4 win32) 'win32:color_windowtext))))))

(defmethod draw-cell (grid (column-header column-header-row) column port x
                      &optional graphics-args)
 ;;; Draw the interior of the cell
  ;; Args: x  Left internal coordinate of cell rectangle
  #-debug (declare (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let ((column-width (column-width column)))
   (declare (fixnum column-width))
   (when (= column-width 0) (return-from draw-cell))
   (let ((column-%cache% (column-%cache% column))
         (height (the fixnum (- (gp:port-height column-header) 2)))
         (x (the fixnum (1+ (or x (column-left grid column))))))
          ;(left-frozen-ncols (left-frozen-ncols grid))
          ;(mask-left x)
          ;(mask-width (- column-width 3)))
    (declare (fixnum height x))
    (with-cell-mask (mask-left mask-width)
        grid column x (setq column-width (the fixnum (- column-width 3)))
      (unless port (setq port column-header))
      (when (eq column-%cache% *unbound-value*)
        (setf column-%cache% (wrap-column-text column-header
                                               (column-title column) column
                                               port
                                               column-width ;(- column-width 3)
                                               height ;(the fixnum (- height 2))
                                               (or (getf graphics-args :font)
                                                   (capi:simple-pane-font column-header))
                                               #'princ-to-string)
              (column-%cache% column) column-%cache%))
       ;; Refuse if x falls outside visible area
        ;(unless (and (< 0 (+ x column-width (- (scroll-x grid))))
        ;             (<= (- x (scroll-x grid)) (gp:port-width port)))
        ;  (return-from draw-cell)))
      (with-translated-mask (port nil :left mask-left :top 1
                                      :width mask-width :height height) ;(- height 2))
       ;(with-transformed-mask (port nil :left x :top 1
       ;                                 :width (the fixnum (- column-width 3))
       ;                                 :height (the fixnum (- height 2)))
        (cond ((column-selected column)
               (gp:with-inverse-graphics (port)
                 #1=(draw-wrapped-text column-header column-%cache% port
                                       x 1
                                       column-width ;(the fixnum (- column-width 3))
                                        height ;(the fixnum (- height 2))
                                        (column-header-align column) :center
                                        graphics-args)))
              ((and (enabled column) (enabled grid))
               #1#)
              ((draw-wrapped-text column-header column-%cache% port
                                  x 1
                                  column-width ;(the fixnum (- column-width 3))
                                  height ;(the fixnum (- height 2))
                                  (column-header-align column) :center
                                  (list* :foreground
                                          #-(and lispworks4 win32) :color_graytext
                                          #+(and lispworks4 win32) 'win32:color_graytext
                                         graphics-args))) ))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  MOUSE GESTURES  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declaim (ftype (function (t fixnum &optional fixnum) t) column-at-position))
(defun column-at-position (header visual-x &optional (tolerance 0))
 ;;; Args: visual-x  Visual coordinate within the header
  ;;       tolerance If =0, then we skip hidden columns, otherwise can return the
  ;;		     (leftmost hidden) column to the left of the cursor within tolerance.
  (declare (fixnum visual-x tolerance)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (do* ((grid (capi:element-interface header))
        (columns (columns grid))
        (ncols (ncols grid))
        (left-frozen-ncols (left-frozen-ncols grid))
        (x visual-x) ;(+ visual-x (scroll-x grid)))	; translate to pinboard internal
        (right 0)
        column
        (prev nil)
        (col 0 (1+ col)))
       ((>= col ncols)
        prev)
    (declare (fixnum ncols left-frozen-ncols x right col)
             (type simple-vector columns))
    (let ((column-width (column-width (setq column (svref columns col)))))
      (declare (fixnum column-width))
      (cond (prev 
             (return (if (zerop column-width)
                         column				; the leftmost hidden
                         prev)))
            ((< (if (= col left-frozen-ncols)
                    (incf x (the fixnum (scroll-x grid))); translate to internal
                    x)					; compare to visual
                (incf right column-width))		; __x__I__
             (return column))
            ((< x (the fixnum (+ right tolerance)))	; __I_x_
             (setq prev column)) ))))			; return on the next iteration


(defun column-header-motion (header visual-x y)
  (declare (fixnum visual-x y *header-resize-cursor-tolerance*)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (if (<= 0 y (gp:port-height header))
    (with-accessors ((select-press-column select-press-column)) header
      (do* ((grid (capi:element-interface header))
            (columns (columns grid))
            (ncols (ncols grid))
            (left-frozen-ncols (left-frozen-ncols grid))
            (x visual-x) ;(+ visual-x (scroll-x grid)))	; translate to pinboard internal
            (right 0)
            (col 0 (1+ col)))
           ((>= col ncols)
            (cond ((< visual-x *header-resize-cursor-tolerance*)
                   (column-header-resize-cursor header)	 	; at the row header
                   (setf select-press-column 'row-header))
                  ((ywi:restore-cursor header))))
        (declare (fixnum ncols left-frozen-ncols x right col)
                 (type simple-vector columns))
        (cond ((< (if (= col left-frozen-ncols)
                      (incf x (the fixnum (scroll-x grid)))	; translate to internal
                      x)					; compare to visual
                  (incf right (the fixnum (column-width (svref columns col)))))
               (return						; __x__I__
                (cond ((<= (- right *header-resize-cursor-tolerance*) x)
                       (column-header-resize-cursor header)
                       (nilf select-press-column))
                      ((< visual-x *header-resize-cursor-tolerance*)
                       (column-header-resize-cursor header)	; row header
                       (setf select-press-column 'row-header))
                      ((neq (choice-interaction grid) :no-selection)
                       (column-header-select-cursor header)
                       (nilf select-press-column)))))
              ((< x (+ right *header-resize-cursor-tolerance*))	; __I_x_
               (return (column-header-resize-cursor header))) )))
    (ywi:restore-cursor header)))

(defun column-header-button-1-press (header x y)
  (declare (ignore y))
  (yl:debug-format :grid "column-header-press x=~d" x) 
  (with-slots (resize-press-x) header
    (with-accessors ((select-press-column select-press-column)) header
      (cond ((column-header-resize-cursor-p header)
             (setf resize-press-x x)
             (unless select-press-column
               (setf select-press-column (column-at-position header x
                                                *header-resize-cursor-tolerance*))))
            ((column-header-select-cursor-p header) 
             (when-let (column (column-at-position header x))
               (let ((grid (capi:element-interface header))
                     (header-click-callback (slot-value column 'header-click-callback)))
                 (cond (header-click-callback
                        (funcall header-click-callback grid column))
                       (t 
                        ;; a la :extended-selection
                        (deselect-section grid :all)
                        (select-section grid column t)
                        (setf select-press-column column)
                   )))))
            (t ;; else no column under cursor
             (nilf resize-press-x select-press-column))))))

(defun column-header-control-button-1-press (header x y)
  (declare (ignore y))
  (when (column-header-select-cursor-p header) 
    (with-slots (resize-press-x) header
      (with-accessors ((select-press-column select-press-column)) header
        (when-let (column (column-at-position header x))
          (let ((grid (capi:element-interface header)))
            ;; a la :extended-selection
            (cond ((column-selected column)
                   (deselect-section grid column t)
                   (setf select-press-column column))
                  ((select-section grid column t)
                   (setf select-press-column column)))))
        (nilf resize-press-x)))))

(defun column-header-shift-button-1-press (header x y)
  (declare (ignore y))
  (when (column-header-select-cursor-p header) 
    (with-slots (resize-press-x) header
      (when-let (column (column-at-position header x))
        (let* ((grid (capi:element-interface header))
               (min (min (column-number column) (current-col grid)))
               (max (max (column-number column) (current-col grid))))
          ;; a la :extended-selection       
          (and (loop for c across (columns grid)
                     if (<= min (column-number c) max)
                     unless (eq c column)
                       unless (column-selected c)
                         unless (select-section grid c nil)
                           return nil
                         end
                       end
                     end
                     ;else if (column-selected c)
                     ;     do (deselect-section grid c)
                     finally (return t))
               (select-section grid column t)
               (setf (select-press-column header) column)))
        (setf resize-press-x nil)))))

(defun column-header-button-1-motion (header x y)
  (declare (fixnum x) (ignore y)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (when (column-header-select-cursor-p header)
    (with-accessors ((select-press-column select-press-column)) header
      (let ((column (column-at-position header x)))
        ;(when (and column (not (eql column select-press-column)))
          ;(select-section (capi:element-interface header) column))
        (setf select-press-column column)))))

(defun column-header-button-1-release (header x y)
 ;;; CAUTION: Is not always invoked if the focus is switched from the pinboard
  ;;	      to the editing pane or back.
  ;; REMEDY: Pass :reposition nil to pinboard-redraw!
  (declare (fixnum x) (ignore y))
  (with-slots (resize-press-x) header
    (let ((grid (capi:element-interface header)))
      (with-accessors ((select-press-column select-press-column)) header
        (yl:debug-format :grid "column-header-release x=~d press-x=~s ~s"
                         x resize-press-x select-press-column)
        (cond ((column-header-resize-cursor-p header)
               (when resize-press-x
                 (cond ((eq select-press-column 'row-header)
                        (let ((new-width (+ (gp:port-width (row-header grid))
                                            (- x resize-press-x))))
                          (nilf resize-press-x select-press-column)
                          (%set-row-header-width new-width grid)))
                       (select-press-column
                        (let* ((column select-press-column)
                               (new-width (+ (column-width column)
                                             (- x resize-press-x))))
                          (nilf resize-press-x select-press-column)
                          (%set-column-width new-width grid column t))) )))
              (t
               (nilf select-press-column)
               ;(reposition-panes (pinboard grid))
               (capi:set-pane-focus (pinboard grid))	; -> pinboard-on-focus
    ) ) )     )
    (ywi:restore-cursor header)))    ;(column-header-motion header x y)

(defmethod component-button-1-second-press ((header column-header-row) x y)
 ;;; Set "bets fit" width as the maxumum with of data over all filtered records.
  ;; If header is visible, take the title into acount
  (declare (ignore y))
  (when-let (column (column-at-position header x))
    (let* ((grid (capi:element-interface header))
           (right (+ (column-left grid column) (column-width column)
                     (- (scroll-x grid)))))
      (when (<= (- right *column-header-second-press-tolerance*) x right)
        (setf (column-width grid column) :best-width)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  INTERNALS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun %set-column-width (width grid column &optional interactive)
  ;; Adjust the widths of cells in the column col
  (declare (ignore interactive))
  (with-slots (records internal-width left-frozen-width pinboard) grid
    (let* ((current-column (current-column grid))
           (shown (and current-column (not (column-hidden current-column)))))
      (setq width (column-width-to-pixels width grid column))
      (nilf internal-width left-frozen-width)
      (loop with col = (column-number column)
            for record across records
            do (%invalidate-record-cache record col))
      (setf (slot-value column 'width) width)
      (setf (column-%cache% column) *unbound-value*)
      (column-header-redraw grid)
      (%reset-bars grid :vertical nil)
      ;(capi:display-tooltip pinboard)			; remove tooltip if any
      (pinboard-redraw grid)
      (when (and current-column shown (column-hidden current-column))
        (pinboard-on-move pinboard nil nil
                          (load-time-value (SYS::CHARACTER-TO-GESTURE-SPEC #\Tab)))))))

(defun %set-column-header-height (height grid)
  (with-slots (column-header row-header) grid
    (loop for column across (columns grid)
          do (setf (column-%cache% column) *unbound-value*))
    (setq height (max height 0))
    (capi:set-geometric-hint column-header :visible-min-height height)
    (column-header-redraw grid)
    (row-header-on-display row-header)))

(defmethod select-section ((grid ygrid) (column column) &optional cell-p)
  (unless (column-selected column)
    (let ((record (current-record grid))
          (col (column-number column)))
      (when (and (cell-on-exit grid record (current-column grid))
                 (record-on-exit grid record))
        (when (and cell-p record)			; don't ensure visible nor focus
          (move-cell grid (current-row grid) col nil))	; i.e. hide
        (update-navigation-toolbar grid)
        (setf (column-selected column) t)
        (%invalidate-column grid column :redraw-now :header)
        t))))

(defmethod deselect-section ((grid ygrid) (column column) &optional (cell-p t))
  (when (column-selected column)
    (with-slots (current-col) grid
      (nilf (column-selected column))
      (when (and cell-p (= current-col (column-number column)))
        (when-let (selected (find-if #'column-selected (columns grid)))
          (setf current-col (column-number selected))))
      (%invalidate-column grid column :redraw-now :header))))
