;;;-*- Mode: Lisp; -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Simple property list editor implementation

(defun plist-delete-callback (grid record)
  (remf (slot-value grid 'plist) (grid:cell-value grid record 'name)))

(defun plist-update-callback (grid record)
  (if (getf (grid:record-plist record) :new)
      (lw:nconcf (slot-value grid 'plist) (grid:record-data record))))

(defclass plist-grid (grid:sorted-grid)
 ((plist :initarg :plist :initform () :reader grid-plist))
 (:default-initargs
  :column-descriptions '((name ;:title "Property Name"
                          :data-write-converter intern 
                          :sort-asc string-lessp :sort-desc string-greaterp)
                         (value ;:title "Property Value"
                          :print-function prin1-to-string
                          :data-read-converter prin1-to-string
                          :data-write-converter read-from-string))
  :delete-callback 'plist-delete-callback
  :update-callback 'plist-update-callback))

(defmethod grid:fetch-grid ((grid plist-grid) (plist list) &key)
  (do ((rest (setf (slot-value grid 'plist) plist) (cddr rest)))
      ((null rest))
    (grid:add-record grid rest)))

(defmethod (setf grid-plist) (plist (grid plist-grid))
  (capi:execute-with-interface grid
    (lambda (grid plist)
      (grid:wipe-grid grid :redraw-now t)
      (grid:fetch-grid grid plist))
    grid plist))	;(copy-list plist)?
  

;;; Test interface

(capi:define-interface plist-grid-test ()
 ()
 (:panes
  (grid plist-grid
        :accessor grid:grid
        :title "Property list grid:"
	:title-position :top
        ;:row-header-width 0
        ;:navigation-title nil
        ;:horizontal-scroll nil
        ;:allow-edits nil
 ))
 (:layouts
  (main capi:column-layout '(top-docking grid))
  (top-docking capi:docking-layout '()
	:items ()
	:divider-p t))
 (:menu-bar)
 (:default-initargs
  :title "Property List Grid Test Interface"
  :layout 'main
  :message-area t
  :best-width 400 :best-height 300
  :top-level-hook #'ywi:top-level-hook))

(defmethod initialize-instance :after ((self plist-grid-test) &key &allow-other-keys)
  (let* ((layout (slot-value self 'top-docking))
         (items (capi:docking-layout-items layout)))
    (setf items (yl:insert (ywi:make-edit-toolbar) items)
          items (yl:insert (grid:make-records-toolbar) items)
          (capi:docking-layout-items layout) items))
  (ywi:insert-menu 'ywi:edit-menu self)
  (ywi:insert-menu 'grid:records-menu self))

(defmethod grid:operate-on-grid :after ((interface plist-grid-test) grid operation)
  (lw:when-let (toolbar (ywi:find-pane-by-name (slot-value interface 'top-docking)
                                               'grid:records-toolbar))
    (capi:update-toolbar toolbar)))

#||;;;;;;;;;;;;;;;;;;;;;;;;  Sample transcript ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(setq plist '(a 1 b "22" c 3.14 d abcd)
      test (make-instance 'plist-grid-test)
      g (slot-value test 'grid))
(capi:display test)
(setf (grid-plist g) plist)
(capi:execute-with-interface test
  #'(setf grid-plist) '(aa (1 3) bb (3 4)) (grid:grid test))
(setf (grid-plist g) '(aa (1 3) bb (3 4)))
;(capi:execute-with-interface test #'grid:wipe-grid g :redraw-now t)
(setq plist (grid-plist g))
(capi:destroy test)
||#


