;;;-*- Mode: Lisp; -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Date/time columns with popup editing device

(in-package :ystok.grid)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  DATE-COLUMN  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; A la text-input-choice

(defclass date-column (popup-column-mixin text-column)
 ()
 (:default-initargs
  :align :center
  :width '(:character 12)	; +2 for button?
  :print-function	'ywi:date-princ-to-string
  :data-read-converter	'ywi:date-princ-to-string	;ywi:ts-princ-to-string
  :data-write-converter 'ywi:date-write-converter	;ywi:date-or-ts-write-converter
  :sort-asc 'grid:time< :sort-desc 'grid:time>
  :popup-type 'ywi:calendar-popup
  :image 1))

(defmethod read-displayer-data ((column date-column) &optional record)
 ;;;     ?
  ;(with-slots (ywi::%old-value%) popup
  (let ((grid (capi:element-interface *chooser*))
        (overlay (overlay *chooser*))
        text)
    (if (and overlay (editor:buffer-modified (capi:editor-pane-buffer overlay)))
        (if (zerop (length (setq text (string-right-trim *whitespace-bag*
                                                     (capi:editor-pane-text overlay)))))
            nil
            (let ((*grid* grid)
                  (*record-data* (record-data record))
                  (*column* column))
              (ignore-errors
                (funcall (column-data-write-converter column) text))))
        (cell-value grid record column))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  DATE-INTERVAL-COLUMN  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The following two (Sportclub) do not handle intersections and "s" correctly.
;;; (defun tim-interval-lessp (x y)
;;;   (grid:nullable< (lt:interval-start x) (lt:interval-end y)))
;;; (defun tim-interval-greaterp (x y)
;;;   (grid:nullable> (lt:interval-start x) (lt:interval-end y)))

(defun time-interval-lessp (x y)
 ;;; For convenience of sorting records, not for testing "real" relationship
  ;; of time intervals
  ;; NB: Equivalent to J.F.Allen: {< m o s} but infinity is possible for both start and end
  (cond ((null x) (not (null y)))
        ((null y) nil)
        ((let ((start-x (lt:interval-start x))
               (start-y (lt:interval-start y)))
           ;(or (time< start-x start y)
           (cond ((null start-x) (not (null start-y)))		; oo < start-y
                 ((null start-y) nil)
                 ((or (lt:local-time< start-x start-y)
                      (and (lt:local-time= start-x start-y)
                           (let ((end-x (lt:interval-end x))
                                 (end-y (lt:interval-end y)))
                             (cond ((null end-y) (not (null x)))	; end-x < oo
                                   ((null end-x) nil)
                                   (t (lt:local-time< end-x end-y))))))))))))

(defun time-interval-greaterp (x y)
  (time-interval-lessp y x))

(defun date-interval-column-completion-function (grid record column text)
  (declare (ignore text))
  (let ((face (capi:top-level-interface grid))
        (old (cell-value grid record column)))
    (if-bind (value (ywi:popup-time-range old (unless old
                                                '(:range :month
                                                  :map ((:month . :this))))
                                          :timestamp nil
                                          :title (column-title column)
                                          :owner face))
      (values (first-or-self value) :save)			;   
      (values nil :cancel))))

(defclass date-interval-column (completion-column)
 ()
 (:default-initargs
  :header-align :center
  :width '(:string :title)
  :image 1
  :print-function       'date-interval-princ-to-string
  :data-read-converter	'date-interval-princ-to-string
  :data-write-converter 'date-interval-write-converter
  :completion-function  'date-interval-column-completion-function
  :sort-asc 'time-interval-lessp :sort-desc 'time-interval-greaterp))
