;;;-*- Mode: Lisp; Encoding: (win32:code-page :id 1251) -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Row header facility
;;;
;;; Linux/Mac OS TODO: Provide meaningful resize and select cursors

(in-package :ystok.grid)

(defparameter *default-row-header-width* 25)

;;; Maybe use gp:register-image-translation and gp:load-image instead?
(defconstant +row-header-image-width+ 11 "One image width")
(defvar *row-header-external-image*
  #.(gp:read-external-image (lw:current-pathname "images/rowhdr.bmp")
                            :transparent-color-index 1))

(defvar *row-header-image* nil "Images to be drawn on an ordinary row header")
(defvar *row-header-image-selected* nil "Images to be drawn on a selected row header")

(defun row-header-image (header)
  (if (or (null *row-header-image*) (gp:image-freed-p *row-header-image*))
      (setq *row-header-image* (gp:convert-external-image header
                                                          *row-header-external-image*))
      *row-header-image*))

(defun row-header-image-selected (header)
  (if (or (null *row-header-image-selected*)
          (gp:image-freed-p *row-header-image-selected*))
      (gp:with-graphics-state (header :background
                                       #-(and lispworks4 win32) :color_3dhighlight
                                       #+(and lispworks4 win32) 'win32:color_3dhighlight)
        (setq *row-header-image-selected* (gp:convert-external-image header
                                                        *row-header-external-image*)))
      *row-header-image-selected*))

;;; As on LW4 |LoadCursorFromFile| simply returns a handle, we must provide
;;; a mechanizm for setting up the cursors in a delivered image as well.
;;; We are relying on the fact that it returns 0 if file not found.
;;; NB1: load-time-value - loads from bin subdir!
;;; NB2: For other platforms, add the folliwing
;;;  #+(and ?? (not lispworks4)) (ywi:load-cursor-from-file #.(lw:current-pathname "??"))

(defvar *row-header-resize-cursor*
  #+(or win32 (not lispworks4)) (ywi:load-cursor-from-file
                                 #.(lw:current-pathname "images/ResizeNS.cur"))
  #-(or win32 (not lispworks4)) :v-double-arrow)
(defun row-header-resize-cursor-p (pane)
  (ywi:cursor-eql pane *row-header-resize-cursor*))
(defun row-header-resize-cursor (pane)
  (unless (row-header-resize-cursor-p pane)		; provide resize cursor
    (ywi:set-cursor pane *row-header-resize-cursor*)))

(defvar *row-header-select-cursor*
  #+(or win32 (not lispworks4)) (ywi:load-cursor-from-file
                                 #.(lw:current-pathname "images/lrarrow.cur"))
  #-(or win32 (not lispworks4)) :hand)
(defun row-header-select-cursor-p (pane)
  (ywi:cursor-eql pane *row-header-select-cursor*))
(defun row-header-select-cursor (pane)
  (unless (row-header-select-cursor-p pane)	 	; provide select cursor
    (ywi:set-cursor pane *row-header-select-cursor*)))

#+(and win32 lispworks4)
(lw:define-action "Delivery Actions" "Drop load-time grid cursors"
   #'(lambda ()
       (declare (special *column-header-select-cursor* *column-header-resize-cursor*))
       (yl:allf 0
         *column-header-select-cursor* *column-header-resize-cursor*
         *row-header-select-cursor* *row-header-resize-cursor*)))

(defparameter *row-header-input-model*
  (nconc
   '((:motion			    row-header-motion)
     ((:motion :control)	    row-header-motion)
     ((:motion :shift)		    row-header-motion)
     ((:button-1 :motion)	    row-header-button-1-motion)
     ((:button-1 :motion :control)  row-header-button-1-motion)
     ((:button-1 :motion :shift)    row-header-button-1-motion)
     ((:button-1 :press)	    row-header-press)		; :button-3 the same
     ((:button-1 :press :control)   row-header-press :control)
     ((:button-1 :press :shift)	    row-header-press :shift)
     ((:button-1 :release)	    row-header-release)
     ((:button-1 :release :control) row-header-release)
     ((:button-1 :release :shift)   row-header-release))
   *component-input-model*))
                  
(defclass row-header-column (grid-component capi:output-pane)
 ((resize-press-y :initform nil)		; button press - PRESSED/DRAGGED for resize
  (press-row :accessor press-row :initform nil)	; button press - PRESSED/DRAGGED for select
  (motion-row :accessor motion-row :initform nil); but-1 motion- PRESSED/DRAGGED for select
  (timer :accessor timer :initform nil)	; for auto-scroll when pointer outside the header
  (header-align :accessor column-header-align
	:initarg :header-align  :initform :center))
 (:default-initargs
  :display-callback 'row-header-on-display
  :input-model *row-header-input-model*
  :accepts-focus-p nil
  :background #-(and lispworks4 win32) :color_3dface
              #+(and lispworks4 win32) 'win32:color_3dface
  :foreground #-(and lispworks4 win32) :color_btntext
              #+(and lispworks4 win32) 'win32:color_btntext
  :visible-min-width *default-row-header-width* :visible-max-width t
  :visible-min-height 1 :visible-max-height nil
  :visible-border nil))
  ;:help-key #L"LEFT-CLICK on this header to select the entire row"))

;;; VIEW
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#+(and win32 lispworks4)
(defmethod component-on-create :after ((header row-header-column))
  (when (zerop *row-header-select-cursor*)
    (setq *row-header-select-cursor* (ywi:load-cursor-from-file
                                      (lw:current-pathname "images/lrarrow.cur"))
          *row-header-resize-cursor* (ywi:load-cursor-from-file
                                      (lw:current-pathname "images/ResizeNS.cur")))))

(defun row-header-on-display (row-header &optional left (top 0) width
                                                   (height (gp:port-height row-header)))
 ;;; Args: left,top,width,height are usually not passed
  (declare (ignore left width) (fixnum top height)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let ((grid (capi:element-interface row-header))
        (bottom (the fixnum (+ top height)))
        (width (gp:port-width row-header))
        (border (border row-header))
        (*background-cleared* t))
    (declare (fixnum bottom width))
    (cond ((= width 0)
           (return-from row-header-on-display))
          ((= width 1)
           (when border
             (gp:draw-line row-header 0 0 0 height
                         :foreground #1= #-(and lispworks4 win32) :color_windowtext
                                         #+(and lispworks4 win32) 'win32:color_windowtext))
           (return-from row-header-on-display)))
    (with-slots (nrows row-height column-header) grid
      (let* ((column-header-height (the fixnum (1- (gp:port-height column-header))))
             (topmost-exposed-row (topmost-exposed-row grid))
             ;; bottom visual coordinate of area filled with rows
             (rows-bottom (+ column-header-height
                            (* (the fixnum (- nrows topmost-exposed-row)) row-height))))
        (declare (fixnum column-header-height topmost-exposed-row))
        (gp:clear-rectangle row-header 0 top width (- rows-bottom top))
        (when (< rows-bottom bottom)      ; fill background outside internal grid area
          (gp:draw-rectangle row-header 0 (max top rows-bottom)
                             width
                             (the fixnum (- height (the fixnum (- rows-bottom top))))
                             :filled t :foreground *workspace-background*))
        ;; If needed, draw top-left corner imitating 3d line
        (let ((top-row                                ; top row being rendered
               (cond ((<= top column-header-height)
                      (decf width)
                      (if border
                          (progn 				;3dshadow?
                            (gp:draw-rectangle row-header 0 0 width column-header-height
                                               :foreground #1#)
                            (gp:draw-line row-header 1 1 width 1
                             :foreground
                               #2= #-(and lispworks4 win32) :color_3dhighlight
                                   #+(and lispworks4 win32) 'win32:color_3dhighlight))
                          (progn
                            (gp:draw-line row-header width 0 width column-header-height
                                          :foreground #1#)
                            (gp:draw-line row-header 0 0 width 0
                                          :foreground #2#)))
                      topmost-exposed-row)
                     ((the fixnum (+ topmost-exposed-row
                                     (values (truncate (- top column-header-height)
                                                       row-height))))))))
          (declare (fixnum top-row))
          (loop for row fixnum upfrom top-row below nrows
                and y upfrom (+ column-header-height
                                (* (the fixnum (- top-row topmost-exposed-row))
                                   row-height))
                      by row-height below bottom
                do (draw-cell grid (%row-record grid row) row-header row-header y))) ))))

(defun row-header-cell-top (grid row)
 ;;; Value: Top visual coordinate of cell rectangle including border
  ;;	    or nil if falls outside visible area.
  (let* ((column-header-height (the fixnum (1- (gp:port-height (column-header grid)))))
         (y (+ column-header-height (* (the fixnum (- row (topmost-exposed-row grid)))
                                       (row-height grid)))))
    (declare (fixnum column-header-height y))
    (if (and (<= column-header-height y) (< y (gp:port-height (row-header grid))))
        y
        nil)))

(defmethod draw-cell (grid record (row-header row-header-column) port y
                      &optional graphics-args)
 ;;; Draw entire cell: interior and border
  ;; Args: y  Top visual coordinate of cell rectangle including border
  #-debug (declare (optimize (speed 3) (safety 0) (debug 0) (hcl:fixnum-safety 0)))
  (let ((width (gp:port-width row-header)))
   (declare (fixnum width))
   (when (<= width 1) (return-from draw-cell))
   (unless port (setq port row-header))
   (let ((row (record-row-number record))
         (row-height (row-height grid))
         (border (border row-header)))
     (declare (fixnum row-height row))
     (or y
         (setq y (row-header-cell-top grid row))
         (return-from draw-cell))
     #|(unless y
       ;; Let the method itself compute y and refuse if falls outside visible area
       (let ((column-header-height (the fixnum (1- (capi:simple-pane-visible-height 
                                                    (column-header grid))))))
         (declare (fixnum column-header-height))
         (setq y (+ column-header-height (* (the fixnum
                                                 (- row (topmost-exposed-row grid)))
                                            row-height)))
         (unless (and (<= column-header-height y)
                      (< y (capi:simple-pane-visible-height row-header)))
           (return-from draw-cell))))|#
     ;; Draw 3d border
     (cond ((record-selected record)
            (gp:draw-rectangle row-header (if border 2 1) (+ y 2)
                               (the fixnum (- width 3))
                               (the fixnum (- row-height 3))
                               :filled t
                               :foreground #1=
                                 #-(and lispworks4 win32) :color_3dhighlight
                                 #+(and lispworks4 win32) 'win32:color_3dhighlight)
            (when border
              (gp:draw-line row-header 0 y 0 (+ y row-height)	;3dshadow?
                         :foreground #2= #-(and lispworks4 win32) :color_windowtext
                                         #+(and lispworks4 win32) 'win32:color_windowtext))
            (gp:draw-polygon row-header (list (the fixnum (1- width))
                                              (the fixnum (1+ y))
                                              (the fixnum (1- width))
                                              (the fixnum (+ y row-height -1))
                                              0 (the fixnum (+ y row-height -1)))
                             :foreground #3= #-(and lispworks4 win32) :color_window
                                             #+(and lispworks4 win32) 'win32:color_window))
           (t
            (unless *background-cleared*
              (gp:clear-rectangle row-header 0 y width row-height))
            (if border
                (gp:draw-rectangle row-header 0 y (the fixnum (1- width)) row-height
                                   :foreground #2#)		;3dshadow
                (gp:draw-polygon row-header (list 0 y
                                                  (the fixnum (1- width)) y
                                                  (the fixnum (1- width))
                                                   (the fixnum (+ y row-height)))
                                                  ;0 (the fixnum (+ y row-height)))
                                 :foreground #2#))		;3dshadow
            (gp:draw-line row-header (the fixnum (if border 1 0)) (the fixnum (1+ y))
                          (the fixnum (- width 2)) (the fixnum (1+ y))
                          :foreground #1#)))

     ;; Choose image to draw if any and center it horizontally.
     (cond ((= row (current-row grid))
            (gp:draw-image row-header
               (if (record-selected record)
                   (row-header-image-selected row-header)
                   (row-header-image row-header))		; /2
               #4=(case (column-header-align row-header)
                    (:left  1)
                    (:right (- width +row-header-image-width+ 1))
                    (otherwise
                     (max (the fixnum (ash (the fixnum (- width
                                                          +row-header-image-width+))
                                           -1)) 1)))
               (the fixnum (+ y 2))
               :from-x (the fixnum (if (%get-edit-mode grid)		; *2
                                       (the fixnum (ash +row-header-image-width+ 1))
                                       0))
               :to-height (the fixnum (- row-height 2))
               :to-width +row-header-image-width+
               :from-width +row-header-image-width+
               :operation (if (record-selected record) boole-orc1 boole-1))) ;boole-eqv
           ((record-auto-p grid row)
            (gp:draw-image row-header
               (if (record-selected record)
                   (row-header-image-selected row-header)
                   (row-header-image row-header))
               #4# (the fixnum (+ y 2))
               :from-x +row-header-image-width+				; *1
               :to-height (the fixnum (- row-height 2))
               :to-width +row-header-image-width+
               :from-width +row-header-image-width+
               :operation (if (record-selected record) boole-orc1 boole-1)))
           ((getf (record-plist record) :deleted)
            (gp:draw-image row-header
               (if (record-selected record)
                   (row-header-image-selected row-header)
                   (row-header-image row-header))
               #4# (the fixnum (+ y 2))
               :from-x (the fixnum (ash +row-header-image-width+ 2))	; *4
               :to-height (the fixnum (- row-height 2))
               :to-width +row-header-image-width+
               :from-width +row-header-image-width+
               :operation (if (record-selected record) boole-orc1 boole-1))) ))))

;;; CONTROLLER
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declaim (ftype (function (t fixnum) fixnum) row-at-position))
(defun row-at-position (row-header y)
 ;;; Find grid row corresponding to visual vertical coordinate y within row-header
  ;; Value: -1 If is out of range
  (declare (fixnum y)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let ((grid (capi:element-interface row-header)))
    (with-slots (column-header) grid
      (let* ((dy (the fixnum (- y (the fixnum (1- (gp:port-height column-header))))))
             (row (the fixnum (+ (topmost-exposed-row grid)
                                 (values (truncate dy (row-height grid)))))))
        (declare (fixnum dy row))
        (if (< -1 row (nrows grid)) row -1)))))

;;; TODO: Draw Finite state machine MOUSE-TRACKER
;;;
;;;   state IDLE:
;;;     on :press execute 'save-mouse-coordinates and goto-state PRESSED
;;;     otherwise ignore
;;;   end state
;;;
;;;   state PRESSED:
;;;     on :release execute 'clicked and goto-state IDLE
;;;     on :second-press execute 'double-click and goto-state IDLE
;;;     on :motion execute 'relative-motion and goto-state DRAGGING
;;;     otherwise ignore
;;;   end state
;;;
;;;   state DRAGGING:
;;;     on :release execute 'finalize-drag and goto-state IDLE
;;;     on :motion execute 'relative-motion and goto-state DRAGGING
;;;     otherwise ignore
;;;   end state
;;;
;;; end state-machine
;;;
;;; NB1: We dispose of (capi:current-pointer-position :relative-to header)
;;; NB2: On the contrast to list-panel
;;;      1) postpone changing the current-row until the button release event,
;;;      2) set the current-row to the first row on the press event.

(defun row-header-motion (header x y)
 ;;; Only set/reset cursor, modifier does not matter
  (declare (fixnum x y *header-resize-cursor-tolerance*)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let* ((grid (capi:element-interface header))
         (column-header-height (gp:port-height (column-header grid))))
    (declare (fixnum column-header-height))
    (cond ((or (> y (+ column-header-height (internal-height grid)	; below bottom cell
                       (- (scroll-y grid))))
               (< y (the fixnum (- column-header-height			; above top cell
                                   *header-resize-cursor-tolerance*)))
               (< x 0)						; to the left
               (< (the fixnum (gp:port-width header)) x))		; to the right
                   ;(< (the fixnum (- (gp:port-width header) x))
                   ;   *header-resize-cursor-tolerance*))
           (ywi:restore-cursor header))
          ((and ;(> y (+ column-header-height *header-resize-cursor-tolerance*))
                (<= (the fixnum (abs (nth-value 1 (round
                                                   (the fixnum (- y column-header-height))
                                                   (the fixnum (row-height grid))))))
                    *header-resize-cursor-tolerance*))
           (row-header-resize-cursor header))
          ((eq (choice-interaction grid) :no-selection)
           (ywi:restore-cursor header))
          (t (row-header-select-cursor header)))))

(defun row-header-press (header x y &optional modifier)
 ;;; Works similarly for button-1 and button-3
  (declare (ignore x))
  ;(yl:debug-format :grid "row-header-press (x,y)=(~d,~d)" x y) 
  (with-slots (resize-press-y) header
   (setf (motion-row header) nil)
   (let ((grid (capi:element-interface header)))
    (cond ((row-header-resize-cursor-p header)
           (setf resize-press-y y
                 (press-row header) nil))
          ((row-header-select-cursor-p header) 
           (let ((row (row-at-position header y)))
            (declare (fixnum row))
            (if (<= 0 row)
                (let ((record (%row-record grid row)))
                  (flet ((%single ()
                           (deselect-section grid (if (eq modifier :control) :rows :all))
                           (setf (press-row header)
                                 (if (select-section grid record nil) row nil)))
                         (%toggle ()
                           (if (record-selected record)
                               (progn (deselect-section grid record nil)
                                 (setf (press-row header) (first (selected-rows grid))))
                               ;; Move current only if none has been selected earlier
                               (let ((selected-rows (selected-rows grid)))
                                 (setf (press-row header)
                                       (if (and (select-section grid record nil)
                                                (not selected-rows))
                                           row
                                           nil))))))
                    (case (choice-interaction grid)
                      (:single-selection
                       (%single))
                      (:multiple-selection
                       (%toggle))
                      (:extended-selection
                       (case modifier
                         (:control
                          (%toggle))
                         (:shift				; deselect all rows on
                          (let ((current-row (current-row grid))); the other side of
                            (deselect-section grid :rows)	; the current-row
                            (select-section grid (cons (setf (press-row header)current-row)
                                                       row))))
                         (otherwise
                          (%single)) ))
                      (otherwise				; :no-selection
                       #1=(setf (press-row header) nil)) )) )
                #1#)				                ; no row under pointer
            (setf resize-press-y nil)))				; not in resize mode
          ((<= y (gp:port-height (column-header grid)))		; top-left cornercl
           (select-section grid :all)) ))))

;;; LW6 Documentation says:
;;;   Timers run in unpredictable threads, actually in the process that is current
;;;   when the scheduled time is reached. This is likely to be the Idle Process
;;;   in cases where LispWorks is sleeping, but it is inherently unpredictable.
;;;
;;; IMPLICATION:
;;;  A timer can be scheduled/unscheduled in any proccess and ywi:aux-funcall unneeded.
;;;
;;; Args: pane CAPI object with the timer accessor.

(defun periodic-apply (pane msec &optional function &rest args)
 ;;; Args: function If NIL, the timer must exist and we are only rescheduleng it
  ;; Value: Timer (instance of mp::time-event strucure).
  (declare (fixnum msec))
  (let ((timer (timer pane)))
    (cond ((null timer)
           (setf timer (apply #'mp:make-timer 'capi:apply-in-pane-process pane
                              function args)
                 (timer pane) timer))
          (function						; replace function and args
           (periodic-stop pane (lambda (timer) (rplacd (slot-value timer 'mp::args)
                                                       (cons function args))))))
          ;(t (setq delay nil))) ; signals - wrong doc: don't reschedule, only set repeat
    (yl:debug-format :grid "periodic-apply ~s msec ~d" pane msec)
    (mp:schedule-timer-relative-milliseconds timer		; relative msec
                                             (max 10 (min 100 (- msec 100)))
                                             (max 10 msec))))	; repeat msec
    
(defun periodic-stop (pane &optional reinitialize)
 ;;; Args: reinitialize If non-NIL, should be a function of signature: (timer);
  ;;			it is called just after unscheduling.
  (when-let (timer (timer pane))
    (unless (mp:timer-expired-p timer)
      (mp:unschedule-timer timer)
      (yl:debug-format :grid "periodic-stop ~s~:[~; reinit~]" pane reinitialize))
    (when reinitialize
      (funcall reinitialize timer))))

(defun periodic-p (pane)
 ;;; True if the pane timer is scheduled
  (when-let (timer (timer pane))
    (not (mp:timer-expired-p timer))))

(defun periodic-pixels-msec (pixels)
 ;;; Timer interval duration in milliseconds
  ;; Args: pixels Coordinate difference between mouse pointer and pane border.
  (min (truncate 4000 pixels) 500))

#-lispworks-personal-edition
(lw:define-action "Delivery Actions" "Keep timer structure"
  (lambda () (pushnew 'mp::time-event			; for (slot-value timer 'mp::args)
                      (lw:delivery-value :symbols-to-keep-structure-info))))


(defun row-header-button-1-motion (header x y)
 ;;; Args: y Can lay outside the port as mouse capture has occured.
  ;; NB1: We can deselect any row except press-row
  ;; NB1: We must hide editing panes beforehand and pass :reposition nil to capi:scroll.
  ;;      Otherwise, if we hide during the scroll, that releases mouse capture and
  ;;	  neither :button-1 :motion nor :button-1 :release event go to the header.
  (declare (type row-header-column header)
           (fixnum y)
           #-debug (ignore x)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (when-let* ((press-row (press-row header))
              (motion-row (or (motion-row header)
                              (and (row-header-select-cursor-p header) press-row))))
   (locally (declare (fixnum press-row motion-row))
    (let* ((grid (capi:element-interface header))
           (top (gp:port-height (column-header grid)))
           (bottom (gp:port-height header))
           (row 0))
      (declare (fixnum top bottom row))
      (when (periodic-p header)					; timer already scheduled
        (if (<= top y bottom)					; if back into pane area,
            (periodic-stop header)				; stop auto-scroll
            (return-from row-header-button-1-motion		; change interval only
              (periodic-apply header (periodic-pixels-msec
                                      (if (< y top) (- top y) (- y bottom)))))))
      (setq row (row-at-position header y))
      (yl:debug-format :grid "rh-button-motion (x,y)=(~d,~d) m ~d r ~d" x y motion-row row)
      (flet ((%down-section (from to)				; section to (de)select
               (declare (fixnum from to))			; when moving down
               (cond ((< from to) (cons from to))
                     ((= from to) from)))
             (%up-section (from to)				; section to (de)select
               (declare (fixnum from to))			; when moving up
               (cond ((> from to) (cons from to))
                     ((= from to) from))) )
        (cond ((minusp row))					; no row under the pointer
              ((< motion-row row)				; Move down
               (cond ((<= y bottom)
                      #1=(when-let (section (%down-section motion-row
                                                           (1- (min press-row row))))
                           (deselect-section grid section))
                      #2=(when-let (section (%down-section (1+ (max press-row motion-row))
                                                           row))
                           (select-section grid section))
                      (setf (slot-value header 'motion-row) row))
                     (t						;    
                      (setq row (bottom-exposed-row grid))	;   
                      #1#
                      #2#
                      (setf (slot-value header 'motion-row) row)
                      (reposition-panes (pinboard grid) :hide)	; hide editing panes
                      (periodic-apply header (periodic-pixels-msec (- y bottom))
                       (lambda (pinboard max)
                         (declare (fixnum max))
                         (let ((motion-row (motion-row header))
                               (press-row (press-row header)))
                           (when (and motion-row (< motion-row max))
                             (if (< motion-row press-row)	
                                 ;; Light-weight deselect
                                 (let ((record (%row-record grid motion-row)))
                                   (when (record-selected record)
                                     (deletef (selected-rows grid) motion-row :count 1)
                                     (setf (record-selected record) nil))
                                   (incf motion-row))
	                         ;; Light-weight select
                                 (let ((record (%row-record grid (incf motion-row))))
                                   (unless (record-selected record)
                                     (setf (record-selected record) t)
                                     (push motion-row (selected-rows grid)))))
                             (setf (motion-row header) motion-row)
                             (capi:scroll pinboard :vertical :step 1 :reposition nil))))
                             #|(if (< motion-row press-row)
                                 (deselect-section grid motion-row)
                                 (select-section grid (1+ motion-row)))
                             (setf (motion-row header) (1+ motion-row)))))|#
                       (pinboard grid) (1- (nrows grid))))))
              ((< row motion-row)				; Move up
               (cond ((<= top y)
                      #5=(when-let (section (%up-section motion-row
                                                         (1+ (max press-row row))))
                           (deselect-section grid section))
                      #6=(when-let (section (%up-section (1- (min press-row motion-row))
                                                         row))
                           (select-section grid section))
                      (setf (slot-value header 'motion-row) row))
                     (t						;     -
                      (setf row (topmost-exposed-row grid))	;   
                      #5#
                      #6#
                      (setf (slot-value header 'motion-row) row)
                      (reposition-panes (pinboard grid) :hide)	; hide editing panes
                      (periodic-apply header (periodic-pixels-msec (- top y))
                       (lambda (pinboard)
                         (let ((motion-row (motion-row header))
                               (press-row (press-row header)))
                           (when (and motion-row (< 0 motion-row))
                             (if (< press-row motion-row)
                                 ;; Light-weight deselect
                                 (let ((record (%row-record grid motion-row)))
                                   (when (record-selected record)
                                     (deletef (selected-rows grid) motion-row :count 1)
                                     (setf (record-selected record) nil))
                                   (decf motion-row))
	                         ;; Light-weight select
                                 (let ((record (%row-record grid (decf motion-row))))
                                   (unless (record-selected record)
                                     (setf (record-selected record) t)
                                     (push motion-row (selected-rows grid)))))
                               (setf (motion-row header) motion-row)
                               (capi:scroll pinboard :vertical :step -1 :reposition nil))))
                               #|(if (< press-row motion-row)
                                   (deselect-section grid motion-row)
                                   (select-section grid (1- motion-row)))
                               (setf (motion-row header) (1- motion-row)))))|#
                       (pinboard grid))))) ))))))

(defun row-header-release (header x y &optional button)
 ;;; Change current row only on releasing the button
  ;; TODO: Pop-up a menu if button equals :button-3.
  (declare (ignore button))
  (with-slots (resize-press-y) header
    (let ((grid (capi:element-interface header))
          (press-row (press-row header))
          value)
      (yl:debug-format :grid "rh-release (x,y)=(~d,~d) rpy ~s p ~s m ~s"
                      x y resize-press-y press-row (motion-row header))
      (cond ((and (row-header-resize-cursor-p header)
                  (setq value resize-press-y))
             (setf resize-press-y nil)
             (let ((column-header-height (gp:port-height (column-header grid))))
               (if (<= (abs (- column-header-height value))
                       *header-resize-cursor-tolerance*)
                   (%set-column-header-height (+ column-header-height (- y value)) grid)
                   (%set-row-height (+ (row-height grid) (- y value)) grid))))
            (t
             (if press-row	  			; don't ensure visible nor focus
                 (move-cell grid press-row (current-col grid) nil)	; maybe hide
                 (reposition-panes (pinboard grid)))
             (capi:set-pane-focus (pinboard grid))))	; -> pinboard-on-focus
      (periodic-stop header)
      (nilf (press-row header) (motion-row header))
      (row-header-motion header x y)			; set/reset cursor
      (update-navigation-toolbar grid))))

;;; Resize rows

(defun %set-row-height (height grid)
 ;;; Change all rows height so that the current topmost row keeps being displyed
  ;; Args: height  Includes 1 for horizontal gridline if any.
  ;; Value: Actual height in pixels if was applied.
  (with-slots (nrows pinboard row-header row-height records scroll-y) grid
    (unless (eql (setq height (max (row-height-to-pixels height grid) 1)) row-height)
      (when t 
        ;; TODO: try to figure out if the height has been altered considerably?
        (map nil '%invalidate-record-cache records))
      (setf row-height height)
      (capi:set-vertical-scroll-parameters pinboard
              :max-range (+ (* (+ nrows 2) height))	; surplus of two rows
              :step-size height)
      (let ((new-scroll-y (* (topmost-exposed-row grid) row-height)))
        (cond ((= scroll-y new-scroll-y)
               (unless (%reset-bars grid)
                 (when (plusp (length (columns grid)))	; as may be called from 
                   (pinboard-redraw grid))		; restore-preferences
                 (row-header-on-display row-header)))
              (t (capi:scroll pinboard :vertical :move new-scroll-y)
                 (%reset-bars grid))))
      height)))

(defun %set-row-header-width (width grid)
  (with-accessors ((row-header row-header)) grid
    (capi:set-geometric-hint row-header
         :visible-min-width (max width (if (capi:simple-pane-visible-border grid) 1 0)))
    ;(column-header-redraw grid)
    (row-header-on-display row-header)))

